;------------------------------------------------------------------------------
; transaction logging
;------------------------------------------------------------------------------
; LOG, pointed by *packages-log-top* is a list of transactions
; TRANSACTION is a cons (command . delta), where
; COMMAND is a list of strings: (cmd . contents)
;         CMD ::= "install" | "remove"
; DELTA is a list of elements like COMMAND, that is,
;       list of lists of strings and represents actual changes
;       between previous and current packages state
; transaction example:
; (("install" "openssh-clients") (("install" "openssh" "openssh-clients")))

(define *log-top* (make-cell '()))
(define *log-point* (make-cell (cell-ref *log-top*)))

(define (log-pending)
  (map
   (lambda (e)
     (let ((address (string-append "/packages/pending/" e)))
       (cons e (sort (woo-list-names address) string<))))
   (filter
    (lambda (e)
      (member e '("install" "remove" "upgrade")))
    (woo-list-names "/packages/pending"))))

(define *log-pending* (make-cell (log-pending)))

(define (log-revert-cmd cmd)
  (cond ((string=? cmd "install") "remove")
        ((string=? cmd "remove") "install")
        (else cmd)))

(define (log-revert-command command)
  (cons (log-revert-cmd (car command)) (cdr command)))

(define (log-apply-command command)
  (let ((cmd (car command))
        (contents (cdr command)))
    (map
     (lambda (p)
       (woo-write
        (string-append "/packages/state/" p)
        'state cmd))
     contents)))

(define (log-delta old new)
  (filter-map
   (lambda (cmd)
     (let* ((old (or (assoc-ref old cmd) '()))
            (new (or (assoc-ref new cmd) '()))
            (forward (lset-difference equal? new old))
            (backward (lset-difference equal? old new)))
       (or
        (and (not (null? forward))
             (cons (log-revert-cmd cmd) forward))
        (and (not (null? backward))
             (cons cmd backward)))))
   (lset-union equal?
               (map car old)
               (map car new))))

(define (log-point-up top point)
  (or (and (eq? (cdr top) point) top)
      (log-point-up (cdr top) point)))

; reduces log by eliminating complementary commands
(define (log-reduce top)
  (reverse
   (delete-duplicates
    top
    (lambda (x y)
      (let ((x-command (car x))
            (y-command (log-revert-command (car y))))
        (equal? x-command y-command))))))

(define (log-find top command)
  (or (and (or (null? top)
               (list= equal? command (caar top)))
           top)
      (log-find (cdr top) command)))

;------------------------------------------------------------------------------
(define (log-add command)
  (let* ((pending (log-pending))
         (delta (log-delta (cell-ref *log-pending*) pending)))
    (cell-set! *log-pending* pending)
    (cell-set! *log-point* (cons (cons command delta)
                                 (cell-ref *log-point*)))
    (cell-set! *log-top* (cell-ref *log-point*))))

(define (log-undo)
  (or (log-undo-empty?)
      (begin
        (map log-apply-command
             (cdar (cell-ref *log-point*)))
        (cell-set! *log-point* (cdr (cell-ref *log-point*)))
        (cell-set! *log-pending* (log-pending)))))

; reduces log, then finds complementary command;
; if found, replays log, excluding complementary command
; oterwise noop
(define (log-undo-trans command)
  (let* ((rcommand (log-revert-command command))
         (top (log-reduce (cell-ref *log-point*)))
         (point (log-find top rcommand)))
    (if (null? point) #f
        (let ((rcommand (caar point)))
          (woo-write "/packages/control" 'control "quit")
          (map
           (lambda (command)
             (or (eq? rcommand command)
                 (log-apply-command command)))
           (map car top))))))

(define (log-redo)
  (or (log-redo-empty?)
      (begin
        (cell-set! *log-point*
                   (log-point-up (cell-ref *log-top*)
                                 (cell-ref *log-point*)))
        (log-apply-command (caar (cell-ref *log-point*)))
        (cell-set! *log-pending* (log-pending)))))

(define (log-set top)
  (cell-set! *log-top* top)
  (cell-set! *log-point* (cell-ref *log-top*))
  (cell-set! *log-pending* (log-pending)))

(define (log-get)
  (cell-ref *log-top*))

(define (log-reset)
  (cell-set! *log-top* '())
  (cell-set! *log-point* (cell-ref *log-top*))
  (cell-set! *log-pending* (log-pending)))

(define (log-undo-empty?)
  (null? (cell-ref *log-point*)))

(define (log-redo-empty?)
  (or (null? (cell-ref *log-top*))
      (eq? (cell-ref *log-point*) (cell-ref *log-top*))))

(define (log-pending-empty?)
  (null? (cell-ref *log-pending*)))
